import datetime
import os

import importlib_metadata
import jinja2 as jj
from systemrdl.node import RegNode

from . import TEMPLATE_DIR
from .loader import SimpleLoader


# Tests
def is_single_field(node: RegNode) -> bool:
    """Jinja test to decide if a register contains only a single field.

    Args:
        node (RegNode): Node to be checked.

    Returns:
        bool: returns whether the node contains a single field.
    """
    return len(node.inst.children) == 1


def is_filled_by_field(node: RegNode) -> bool:
    """Jinja test to decide if a register is filled by a single field.

    Args:
        node (RegNode): Node to be checked.

    Returns:
        bool: returns whether the node is filled entirly by a single field.
    """
    return is_single_field(node) and node.inst.children[0].width == get_field_width(
        node
    )


def is_write_only(node: RegNode) -> bool:
    """Jinja test to decide if a register is software write only.

    Args:
        node (RegNode): Node to be checked.

    Returns:
        bool: returns whether the register is write only.
    """
    return node.has_sw_writable and not node.has_sw_readable


def is_read_only(node: RegNode) -> bool:
    """Jinja test to decide if a register is software read only.

    Args:
        node (RegNode): Node to be checked.

    Returns:
        bool: returns whether the register is read only.
    """
    return node.has_sw_readable and not node.has_sw_writable


# Filters
def get_field_width(node: RegNode) -> int:
    """Filter to get the width of a Reg.

    Args:
        node (RegNode): Reg for which to return the width.

    Returns:
        int: width of register in bits.
    """
    return node.size * 8


def get_native_typesize(bit_count: int) -> int:
    """Filter to get native type size in bits that fits the required amount of bits.

    Args:
        bit_count (int): Number of bits for which to return the native type size.

    Raises:
        NotImplementedError: Raises this error if bitcount > 64.

    Returns:
        int: Native type size that fits the amounts of bits.
    """

    if bit_count <= 8:
        return 8
    elif bit_count <= 16:
        return 16
    elif bit_count <= 32:
        return 32
    elif bit_count <= 64:
        return 64
    else:
        raise NotImplementedError(
            "no bit counts greater than 64 implementet in get_native_typesize"
        )


def get_alt_name(name: str) -> str:
    """Filter to obtain an alternative name for the supplied string.

    Args:
        name (str): Input name

    Returns:
        str: Alternative name if there is one, else input
    """
    name_mapping = {
        "Apb3AClint": "aclint",
        "Apb3Gpio": "gpio",
        "Apb3IrqCtrl": "irq",
        "Apb3SpiXdrMasterCtrl": "spi",
        "Apb3Timer": "timer",
        "Apb3Uart": "uart",
    }
    return name_mapping[name] if name in name_mapping else name


class PaddingHolder:
    padding_counter = 0


def get_padding(node: RegNode) -> int:
    """Jinja test to generate the needed amount of bytes for filling gaps to the next regfile.

    Args:
        node (RegNode): Node to be checked.

    Returns:
        int: Amounts of bytes of padding needed
    """
    bytes_to_fill = 0
    neighbors = [child for child in node.parent.children()]
    neighbors = sorted(neighbors, key=lambda x: x.raw_absolute_address)
    index = neighbors.index(node)
    if index < len(neighbors) - 1:
        ends_at = node.raw_absolute_address + node.size
        next_start = neighbors[index + 1].raw_absolute_address
        bytes_to_fill = next_start - ends_at
    return bytes_to_fill


# Globals
def get_padding_counter() -> int:
    """Function that returns the current padding counter and increases it

    Returns:
        int: Current padding counter value
    """
    val = PaddingHolder.padding_counter
    PaddingHolder.padding_counter += 1
    return val


def register_helpers(jj_env: jj.Environment) -> None:
    """Prepares the jinja environment by registering all tests and filters to be available inside
    the templates.

    Args:
        jj_env (jj.Environment): The Environment for which to register the tests and filters.
    """
    jj_env.tests["single_field"] = is_single_field
    jj_env.tests["single_field_filled"] = is_filled_by_field
    jj_env.tests["readonly"] = is_read_only
    jj_env.tests["writeonly"] = is_write_only

    jj_env.filters["width"] = get_field_width
    jj_env.filters["native_size"] = get_native_typesize
    jj_env.filters["alt_name"] = get_alt_name
    jj_env.filters["padding"] = get_padding

    jj_env.globals["now"] = datetime.datetime.now().replace(microsecond=0)
    # This is the solution for using python >3.11 since the line above is then deprecated
    #    jj_env.globals["now"] = datetime.datetime.now(datetime.UTC).replace(microsecond=0)

    jj_env.globals["version"] = importlib_metadata.version("peakrdl-mnrs")
    jj_env.globals["padding_counter"] = get_padding_counter


class OutputFileGenerator:
    """Generator that fills the template."""

    def __init__(self) -> None:
        loader = SimpleLoader([TEMPLATE_DIR, os.getcwd()])
        jj_env = jj.Environment(loader=loader)
        register_helpers(jj_env)
        self.jj_env = jj_env

    def run(
        self,
        template_name: str,
        context: dict,
        output_filename: str,
        overwrite: bool = False,
        never_overwrite: bool = False,
    ) -> None:
        """Fetches and fills the requested template with the given name.

        Only generates a new file if the desired file (content) does not already exist.
        This behavior can be overridden by using the overwrite option.

        Also offers never_overwrite, which will not allow any new file creation if a file with
        the desired output name already exists.

        Args:
            template_name (str): Name of the template in the 'templates' directory or
            (relative) path to custom template.
            context (dict): Available context inside the template.
            output_filename (str): Name of the file that will be created.
            overwrite (bool): Force overwrite if files already exist.
            never_overwrite (bool): Protection against overwrites.
        """
        if os.path.isfile(output_filename):
            if never_overwrite:
                return
            if not overwrite:
                gen_content = self.jj_env.get_template(template_name).render(context)
                with open(output_filename, "r", encoding="utf8") as existing_file:
                    existing_content = existing_file.read()
                    if gen_content == existing_content:
                        return
        with open(output_filename, "w", encoding="utf8") as f:
            PaddingHolder.padding_counter = 0
            self.jj_env.get_template(template_name).stream(context).dump(f)
